/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.exceptions.generic.CertificateValidationException;

import java.security.GeneralSecurityException;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

public class CertificateVerificationUtils {
    public static void assertThatCertificateIsValid(X509Certificate sgdCertificate) throws CertificateValidationException {
        try {
            sgdCertificate.checkValidity();
        } catch (GeneralSecurityException e) {
            throw new CertificateValidationException("Error while trying to validate Certificate", e);
        }
    }

    public static void assertThatCertificateHasKvnrOrTelematikId(X509Certificate sgdCertificate) throws CertificateValidationException {
        try {
            if (SgdCryptoUtilities.extractKVNRfromSubjectDN(sgdCertificate.getSubjectDN().getName()) == null) {
                if (SgdCryptoUtilities.getInstance().extractTelematikIdFromCertificate(sgdCertificate).isEmpty()) {
                    throw new CertificateValidationException("Certificate does not have a KVNR or a Telematik ID!");
                }
            }
        } catch (Exception e) {
            if (e instanceof CertificateValidationException) {
                throw e;
            } else {
                throw new CertificateValidationException("Error while trying to validate Certificate", e);
            }
        }
    }

    public static void assertThatCertificateSignatureIsValid(X509Certificate sgdCertificate, List<PublicKey> possibleSigners) throws CertificateValidationException {
        final List<Optional<CertificateValidationException>> checkResultExceptions = possibleSigners.stream()
                .map(publicKey -> tryToVerifySignatureUsingAndSignerAndReturnExceptionWhenFail(sgdCertificate, publicKey))
                .collect(Collectors.toList());

        if (checkResultExceptions.stream()
                .filter(Optional::isEmpty)
                .findAny().isPresent()) {
            return;
        }

        throw checkResultExceptions.stream()
                .findAny()
                .map(Optional::get)
                .orElseGet(() -> new CertificateValidationException("No Public-Keys supplied for certificate validation!"));
    }

    private static Optional<CertificateValidationException> tryToVerifySignatureUsingAndSignerAndReturnExceptionWhenFail(X509Certificate sgdCertificate, PublicKey publicKey) {
        try {
            sgdCertificate.verify(publicKey);
            return Optional.empty();
        } catch (GeneralSecurityException e) {
            return Optional.of(new CertificateValidationException("Error while trying to validate Certificate (Issuer is " + sgdCertificate.getIssuerDN() + ")", e));
        }
    }
}
