/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.SgdInstance;
import de.gematik.ti.sgd.exceptions.generic.CertificateValidationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.security.PublicKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

import static de.gematik.ti.sgd.components.CertificateVerificationUtils.assertThatCertificateIsValid;
import static de.gematik.ti.sgd.components.CertificateVerificationUtils.assertThatCertificateSignatureIsValid;

public class SgdClientUtils {
    private static final Logger logger = LoggerFactory.getLogger(SgdClientUtils.class);

    public static de.gematik.ti.sgd.domain.json.SgdRequest generateGetPublicKeyRequest(X509Certificate certificate) {
        return generateGetPublicKeyRequest(certificate, null);
    }

    public static de.gematik.ti.sgd.domain.json.SgdRequest generateGetPublicKeyRequest(X509Certificate certificate, byte[] ocspResponse) {
        try {
            String ocspResponseEncodedString = (ocspResponse == null) ? "" : SgdCryptoUtilities.getInstance().convertBytesToB64String(ocspResponse);
            return de.gematik.ti.sgd.domain.json.SgdRequest.builder()
                    .certificate(SgdCryptoUtilities.getInstance().convertBytesToB64String(certificate.getEncoded()))
                    .ocspResponse(ocspResponseEncodedString)
                    .command(de.gematik.ti.sgd.domain.json.SgdCommand.GET_PUBLIC_KEY)
                    .build();
        } catch (CertificateEncodingException e) {
            throw new RuntimeException("Unable to serialize given X509-Certificate");
        }
    }

    public static void performGetPublicKeyVerificationAndThrowExceptionWhenFail(de.gematik.ti.sgd.domain.json.GetPublicKeyResponse getPublicKeyResponse, List<PublicKey> possibleSigners, SgdInstance sgdInstance) {
        final byte[] sgdCertificateRawBytes = SgdCryptoUtilities.getInstance().convertB64StringToBytes(getPublicKeyResponse.getCertificate());
        X509Certificate sgdCertificate;
        try {
            sgdCertificate = SgdCryptoUtilities.getInstance().parseX509CertificateFromBytes(sgdCertificateRawBytes);
        } catch (RuntimeException e) {
            throw new de.gematik.ti.sgd.exceptions.client.InvalidCertificateException("Error while validating SGD Certificate in Client for getPublicKeyResponse", e);
        }

        assertThatCertificateHasAdmissionExtensionOid(sgdCertificate, sgdInstance == SgdInstance.SGD_1 ?
                SgdCryptoUtilities.OID_SGD1_HSM : SgdCryptoUtilities.OID_SGD2_HSM);

        try {
            assertThatCertificateIsValid(sgdCertificate);
            assertThatCertificateSignatureIsValid(sgdCertificate, possibleSigners);
        } catch (CertificateValidationException e) {
            throw new de.gematik.ti.sgd.exceptions.client.InvalidCertificateException("Error while validating SGD Certificate in Client for getPublicKeyResponse (Issued by " + sgdCertificate.getIssuerDN().getName() + ")", e);
        }

        assertThatSignatureInResponseIsValid(sgdCertificate.getPublicKey(), getPublicKeyResponse);
    }

    private static void assertThatSignatureInResponseIsValid(PublicKey signerPublicKey, de.gematik.ti.sgd.domain.json.GetPublicKeyResponse getPublicKeyResponse) {
        try {
            SgdCryptoUtilities.getInstance().verifyEcdsaSignatureAndThrowExceptionWhenFail(getPublicKeyResponse.getPublicKeyEcies().getBytes(),
                    signerPublicKey,
                    SgdCryptoUtilities.getInstance().convertB64StringToBytes(getPublicKeyResponse.getSignature()));
        } catch (Exception e) {
            throw new de.gematik.ti.sgd.exceptions.client.InvalidSignatureException("Error while validating Signature of ECIES-Key in getPublicKeyResponse", e);
        }
    }

    private static void assertThatCertificateHasAdmissionExtensionOid(X509Certificate certificate, String policyOid) {
        if (!SgdCryptoUtilities.getInstance().doesCertificateContainAdmissionsExtensionOid(certificate, policyOid)) {
            throw new de.gematik.ti.sgd.exceptions.client.InvalidCertificateException("Unable to find OID " + policyOid + " in Admission-Extension in getPublicKeyResponse!");
        }
    }

    public static List<PublicKey> extractPublicKeysFromListOfCertificateFilenames(List<String> possibleSignerCertificateFiles) {
        if (possibleSignerCertificateFiles == null) {
            throw new NullPointerException("Given list of possibleSignerCertificateFiles is null. Failing early (Did you forget to configure the signer list?)");
        }
        return possibleSignerCertificateFiles.stream()
                .map(filename -> {
                    try {
                        return Optional.ofNullable(SgdCryptoUtilities.getInstance().loadX509CertificateFromCrtFile(filename));
                    } catch (Exception e) {
                        logger.warn("Error while loading Certificate: ", e);
                        return Optional.empty();
                    }
                })
                .filter(Optional::isPresent)
                .map(Optional::get)
                .map(c -> ((Certificate) c).getPublicKey())
                .collect(Collectors.toList());
    }

    public static List<X509Certificate> extractCertificatesFromListOfFilenames(List<String> possibleOcspSignerCertificateFiles) {
        if (possibleOcspSignerCertificateFiles == null) {
            throw new NullPointerException("Given list of possibleOcspSignerCertificateFiles is null. Failing early (Did you forget to configure the signer list?)");
        }
        return possibleOcspSignerCertificateFiles.stream()
                .map(filename -> {
                    try {
                        return SgdCryptoUtilities.getInstance().loadX509CertificateFromCrtFile(filename);
                    } catch (Exception e) {
                        return null;
                    }
                })
                .filter(Objects::nonNull)
                .collect(Collectors.toList());
    }
}
