/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.domain;

import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ws.fa.phr.v1.KeyType;
import de.gematik.ws.fa.phr.v1.PHRKeyType;
import de.gematik.ws.fd.phrs.authorizationservice.v1.EncryptedKeyContainerType;

import javax.xml.bind.DatatypeConverter;
import java.util.concurrent.ThreadLocalRandom;

import static de.gematik.ti.sgd.domain.util.PHRUtils.*;

public class RecordAndContextKey {
    private final byte[] recordKeyBytes;
    private final byte[] contextKeyBytes;
    private final String kvnr;

    public RecordAndContextKey(byte[] recordKeyBytes, byte[] contextKeyBytes, String kvnr) {
        this.recordKeyBytes = recordKeyBytes;
        this.contextKeyBytes = contextKeyBytes;
        this.kvnr = kvnr;
    }

    public static RecordAndContextKey fromPHRKeyType(PHRKeyType phrKeyType) {
        return RecordAndContextKeyBuilder.aRecordAndContextKey()
                .withKvnr(phrKeyType.getInsurant())
                .withContextKeyBytes(phrKeyType.getContextKey().getValue())
                .withRecordKeyBytes(phrKeyType.getRecordKey().getValue())
                .build();
    }

    public static RecordAndContextKey newRandomKeysForKvnr(String kvnr) {
        byte[] recordKey = new byte[256 / 8];
        ThreadLocalRandom.current().nextBytes(recordKey);
        byte[] contextKey = new byte[256 / 8];
        ThreadLocalRandom.current().nextBytes(contextKey);
        return RecordAndContextKeyBuilder.aRecordAndContextKey()
                .withContextKeyBytes(contextKey)
                .withRecordKeyBytes(recordKey)
                .withKvnr(kvnr)
                .build();
    }

    private static KeyType generateKeyTypeValue(byte[] keyBytes) {
        KeyType result = new KeyType();
        result.setValue(keyBytes);
        result.setAlgorithm(AES_256_GCM_ALGORITHM_IDENTIFIER);
        return result;
    }

    public byte[] getRecordKeyBytes() {
        return recordKeyBytes;
    }

    public byte[] getContextKeyBytes() {
        return contextKeyBytes;
    }

    public String getKvnr() {
        return kvnr;
    }

    public EncryptedKeyContainerType encipherOnceToContainer(byte[] sgd1Key, byte[] sgd1Vector) {
        byte[] payload = marshall(packToPhrObject(), PHR_KEY_ELEMENT_NAME);

        return encryptContentToContainer(sgd1Key, sgd1Vector, payload, DatatypeConverter.printBase64Binary(sgd1Vector));
    }

    private EncryptedKeyContainerType encryptContentToContainer(byte[] key, byte[] associatedDataForEncryption, byte[] payload, String associatedDataForXml) {
        EncryptedKeyContainerType result = new EncryptedKeyContainerType();
        result.setAssociatedData(associatedDataForXml);
        result.setAlgorithm("http://www.w3.org/2009/xmlenc11#aes256-gcm");
        try {
            result.setCiphertext(SgdCryptoUtilities.getInstance().encryptUsingAesGcm(payload, key, associatedDataForEncryption));
        } catch (Exception e) {
            throw new RuntimeException("Error while encrpyting PHR Record", e);
        }
        return result;
    }

    public PHRKeyType packToPhrObject() {
        PHRKeyType result = new PHRKeyType();
        result.setInsurant(kvnr);
        result.setContextKey(generateKeyTypeValue(contextKeyBytes));
        result.setRecordKey(generateKeyTypeValue(recordKeyBytes));
        return result;
    }

    public EncryptedKeyContainerType encipherTwiceToContainer(AesGcmEncryptionData sgd1AesGcmKey, AesGcmEncryptionData sgd2AesGcmKey) {
        return encipherTwiceToContainer(sgd1AesGcmKey.getAesKey(), sgd1AesGcmKey.getAssociatedData(),
                sgd2AesGcmKey.getAesKey(), sgd2AesGcmKey.getAssociatedData());
    }

    public EncryptedKeyContainerType encipherTwiceToContainer(byte[] sgd1KeyBytes, String sgd1Vector, byte[] sgd2KeyBytes, String sgd2Vector) {
        byte[] initialPayload = marshall(packToPhrObject(), PHR_KEY_ELEMENT_NAME);
        final String associatedDataSgd1 = DatatypeConverter.printBase64Binary(sgd1Vector.getBytes());
        final String associatedDataSgd2 = DatatypeConverter.printBase64Binary(sgd2Vector.getBytes());

        EncryptedKeyContainerType singularEncryptedContainer = encryptContentToContainer(sgd1KeyBytes, sgd1Vector.getBytes(), initialPayload, associatedDataSgd1);

        final byte[] intermediarePayload = marshall(singularEncryptedContainer, ENCRYPTED_KEY_CONTAINER_ELEMENT_NAME);
        EncryptedKeyContainerType doubleEncryptedContainer = encryptContentToContainer(sgd2KeyBytes, (sgd1Vector + sgd2Vector).getBytes(), intermediarePayload, associatedDataSgd1 + " " + associatedDataSgd2);

        return doubleEncryptedContainer;
    }

    public static final class RecordAndContextKeyBuilder {
        private byte[] recordKeyBytes;
        private byte[] contextKeyBytes;
        private String kvnr;

        private RecordAndContextKeyBuilder() {
        }

        public static RecordAndContextKeyBuilder aRecordAndContextKey() {
            return new RecordAndContextKeyBuilder();
        }

        public RecordAndContextKeyBuilder withRecordKeyBytes(byte[] recordKeyBytes) {
            this.recordKeyBytes = recordKeyBytes;
            return this;
        }

        public RecordAndContextKeyBuilder withContextKeyBytes(byte[] contextKeyBytes) {
            this.contextKeyBytes = contextKeyBytes;
            return this;
        }

        public RecordAndContextKeyBuilder withKvnr(String kvnr) {
            this.kvnr = kvnr;
            return this;
        }

        public RecordAndContextKey build() {
            return new RecordAndContextKey(recordKeyBytes, contextKeyBytes, kvnr);
        }
    }
}
