/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.SgdTestCryptoUtilities;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.json.EncryptedMessageContainer;
import de.gematik.ti.sgd.domain.json.GetPublicKeyResponse;
import de.gematik.ti.sgd.domain.json.SgdRequest;
import de.gematik.ti.sgd.domain.json.SgdResponseStatus;
import de.gematik.ti.sgd.domain.sgdCommunication.EncryptedMessage;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.RandomStringUtils;
import org.junit.Before;
import org.junit.Test;
import org.springframework.test.util.ReflectionTestUtils;

import java.io.File;
import java.security.KeyPair;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.List;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import static de.gematik.ti.sgd.components.SgdClientUtils.extractPublicKeysFromListOfCertificateFilenames;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

public class SgdClientGetAuthenticationTokenTest {
    private static final char[] HEX_CHARS = new char[]{'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', 'a', 'b', 'c', 'd', 'e', 'f'};
    private static List<PublicKey> SIGNER_LIST;

    static {
        SgdCryptoUtilities.setInstance(new SgdTestCryptoUtilities());

        SIGNER_LIST = extractPublicKeysFromListOfCertificateFilenames(Arrays.asList(
                "src/test/resources/keys/ca/GEM.KOMP-CA17-TEST-ONLY.crt",
                "src/test/resources/keys/ca/GEM.KOMP-CA17-TEST-ONLY.crt",
                "src/test/resources/keys/ca/GEM.KOMP-CA10-TEST-ONLY.crt",
                "src/test/resources/keys/ca/GEM.EGK-CA10-TEST-ONLY.crt"));
    }

    private final String P12_1_FILENAME = "src/test/resources/keys/egk/109500969_X114428530_c.ch.enc-ecc.p12";
    private final String P12_4_FILENAME = "src/test/resources/keys/egk/109500969_X114428533_c.ch.aut-ecc.p12";
    private final String ECC_AUT_CERTIFICATE_FILE = "src/test/resources/keys/egk/109500969_X114428530_c.ch.aut-ecc.crt";
    private final String ECC_AUT_KEY_FILE = "src/test/resources/keys/egk/109500969_X114428530_c.ch.aut-ecc.p12";

    // System under test
    private SgdClientKeyDerivationMachine sgdClient;
    private SchluesselGenerierungsDienst sgd1;
    private SchluesselGenerierungsDienst sgd2;
    private X509Certificate sgd1_Certificate;
    private X509Certificate sgd2_Certificate;
    private KeyPair sgd1_Key;
    private KeyPair sgd2_Key;
    private KeyPair sgd1EphemeralKeyPair;
    private KeyPair sgd2EphemeralKeyPair;

    private X509Certificate clientCertificate;
    private KeyPair clientKeyPair;

    @Before
    public void initSgd() {
        if (sgdClient == null) {
            sgd1_Certificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File("src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.p12", "00");
            sgd2_Certificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File("src/test/resources/keys/sgd/c.sgd2.hsm-aut-ecc.p12", "00");
            sgd1_Key = SgdCryptoUtilities.getInstance().loadKeypairFromP12File("src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.p12", "00", "00");
            sgd2_Key = SgdCryptoUtilities.getInstance().loadKeypairFromP12File("src/test/resources/keys/sgd/c.sgd2.hsm-aut-ecc.p12", "00", "00");
            sgd1EphemeralKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_1_FILENAME, "00", "00");
            sgd2EphemeralKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_4_FILENAME, "00", "00");
            sgd1 = new SchluesselGenerierungsDienst(sgd1_Certificate, sgd1_Key.getPrivate(), sgd1EphemeralKeyPair);
            sgd2 = new SchluesselGenerierungsDienst(sgd2_Certificate, sgd2_Key.getPrivate(), sgd2EphemeralKeyPair);

            ReflectionTestUtils.setField(sgd1, "possibleCertificateSigners", SIGNER_LIST);
            ReflectionTestUtils.setField(sgd2, "possibleCertificateSigners", SIGNER_LIST);

            clientCertificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromCrtFile(ECC_AUT_CERTIFICATE_FILE);
            clientKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(ECC_AUT_KEY_FILE, "00", "00");
        }

        sgdClient = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, SIGNER_LIST);

        sgdClient.receivePublicKeyResponse_sgd1(sgd1.receive(sgdClient.getPublicKeyRequest()).toGetPublicKeyResponse());
        sgdClient.receivePublicKeyResponse_sgd2(sgd2.receive(sgdClient.getPublicKeyRequest()).toGetPublicKeyResponse());
    }

    @Test
    public void happyPath() {
        // Request
        final SgdRequest authenticationTokenRequestSgd1 = sgdClient.getAuthenticationTokenRequest_sgd1();
        final SgdRequest authenticationTokenRequestSgd2 = sgdClient.getAuthenticationTokenRequest_sgd2();
        // Response
        final EncryptedMessageContainer sgd1AuthenticationToken = sgd1.receive(authenticationTokenRequestSgd1).toEncryptedMessageContainer();
        final EncryptedMessageContainer sgd2AuthenticationToken = sgd2.receive(authenticationTokenRequestSgd2).toEncryptedMessageContainer();
        sgdClient.receiveAuthenticationTokenResponse_sgd1(sgd1AuthenticationToken);
        sgdClient.receiveAuthenticationTokenResponse_sgd2(sgd2AuthenticationToken);
    }

    @Test
    public void authenticationTokenResponseClientChallengeMismatch_shouldThrowException() {
        final EncryptedMessageContainer sgd1AuthenticationToken =
                // rehash: bleibt gleiche Länge, aber gibt anderen Inhalt
                getAuthTokenResponseAndModifyMessagePayload(parts -> parts.set(1, DigestUtils.sha256Hex(parts.get(1))));

        assertThatThrownBy(() -> sgdClient.receiveAuthenticationTokenResponse_sgd1(sgd1AuthenticationToken))
                .isInstanceOf(IllegalArgumentException.class)
                .hasMessageContaining("Client-Challenge");
    }

    @Test
    public void authenticationTokenResponseDoesNotStartWithResponse() {
        final EncryptedMessageContainer sgd1AuthenticationToken = getAuthTokenResponseAndModifyMessagePayload(parts -> parts.set(0, "HierStehtNichtResponse"));

        assertThatThrownBy(() -> sgdClient.receiveAuthenticationTokenResponse_sgd1(sgd1AuthenticationToken))
                .isInstanceOf(RuntimeException.class)
                .hasMessageContaining("Response");
    }

    @Test
    public void authenticationTokenResponseHValueDoesNotMatch() {
        final EncryptedMessageContainer sgd1AuthenticationToken =
                // rehash: bleibt gleiche Länge, aber gibt anderen Inhalt
                getAuthTokenResponseAndModifyMessagePayload(parts -> parts.set(2, DigestUtils.sha256Hex(parts.get(2))));

        assertThatThrownBy(() -> sgdClient.receiveAuthenticationTokenResponse_sgd1(sgd1AuthenticationToken))
                .isInstanceOf(IllegalArgumentException.class)
                .hasMessageContaining("h-Value");
    }

    @Test
    public void authenticationTokenResponse_AuthenticationTokenDoesNotStartWithAT() {
        final EncryptedMessageContainer sgd1AuthenticationToken = getAuthTokenResponseAndModifyMessagePayload(parts -> parts.set(3, parts.get(3).replaceFirst("AT", "BA")));

        assertThatThrownBy(() -> sgdClient.receiveAuthenticationTokenResponse_sgd1(sgd1AuthenticationToken))
                .isInstanceOf(RuntimeException.class)
                .hasMessageContaining("BA");
    }

    @Test
    public void authenticationTokenResponse_AuthenticationTokenNot256Bit() {
        final EncryptedMessageContainer sgd1AuthenticationToken = getAuthTokenResponseAndModifyMessagePayload(parts -> parts.set(3, "AT" + RandomStringUtils.random(60, HEX_CHARS)));

        assertThatThrownBy(() -> sgdClient.receiveAuthenticationTokenResponse_sgd1(sgd1AuthenticationToken))
                .isInstanceOf(RuntimeException.class)
                .hasMessageContaining("Hex-String");
    }

    private EncryptedMessageContainer getAuthTokenResponseAndModifyMessagePayload(Consumer<List<String>> messagePatcher) {
        final EncryptedMessageContainer sgd1AuthenticationToken = sgd1.receive(sgdClient.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer();

        final KeyPair clientEciesKeyPair = (KeyPair) ReflectionTestUtils.getField(sgdClient, "ephemeralEciesKeyPair");
        List<String> payload = Arrays.asList(new String(sgd1AuthenticationToken.getEncryptedMessage().decryptMessage(clientEciesKeyPair))
                .split(" "));

        messagePatcher.accept(payload);

        EncryptedMessageContainer rewrittenContainer = EncryptedMessageContainer.EncryptedMessageContainerBuilder.anEncryptedMessageContainer()
                .withStatus(SgdResponseStatus.OK)
                .withEncryptedMessage(new EncryptedMessage(clientEciesKeyPair.getPublic(), payload.stream()
                        .collect(Collectors.joining(" "))))
                .build();

        return rewrittenContainer;
    }
}